#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.button module

Form buttons management.
"""

import sys

from six import class_types
from zope.interface import Interface, alsoProvides, implementedBy, implementer, providedBy
from zope.interface.adapter import AdapterRegistry
from zope.interface.interfaces import IInterface
from zope.location import Location, locate
from zope.schema import Field, getFieldsInOrder
from zope.schema.fieldproperty import FieldProperty

from pyams_form.action import Action, ActionHandlerBase, Actions
from pyams_form.browser.image import ImageWidget
from pyams_form.browser.submit import SubmitWidget
from pyams_form.interfaces import IValue
from pyams_form.interfaces.button import IActionHandler, IActions, IButton, IButtonAction, \
    IButtonHandler, IButtonHandlers, IButtons, IImageButton
from pyams_form.interfaces.form import IButtonForm, IFormAware, IHandlerForm
from pyams_form.util import SelectionManager, create_id, expand_prefix, get_specification, \
    to_unicode
from pyams_form.value import ComputedValueCreator, StaticValueCreator
from pyams_form.widget import AfterWidgetUpdateEvent
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


# pylint: disable=invalid-name
StaticButtonActionAttribute = StaticValueCreator(
    discriminators=('form', 'request', 'content', 'button', 'manager')
)

# pylint: disable=invalid-name
ComputedButtonActionAttribute = ComputedValueCreator(
    discriminators=('form', 'request', 'content', 'button', 'manager')
)


@implementer(IButton)
class Button(Field):
    """A simple button in a form."""

    access_key = FieldProperty(IButton['access_key'])
    action_factory = FieldProperty(IButton['action_factory'])

    def __init__(self, *args, **kwargs):
        # Provide some shortcut ways to specify the name
        if args:
            kwargs['__name__'] = args[0]
            args = args[1:]
        if 'name' in kwargs:
            kwargs['__name__'] = kwargs['name']
            del kwargs['name']
        # Extract button-specific arguments
        self.access_key = kwargs.pop('access_key', None)
        self.condition = kwargs.pop('condition', None)
        # Initialize the button
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return '<%s %r %r>' %(
            self.__class__.__name__, self.__name__, self.title)


@implementer(IImageButton)
class ImageButton(Button):
    """A simple image button in a form."""

    image = FieldProperty(IImageButton['image'])

    def __init__(self, image, *args, **kwargs):
        self.image = image
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return '<%s %r %r>' %(
            self.__class__.__name__, self.__name__, self.image)


@implementer(IButtons)
class Buttons(SelectionManager):
    """Button manager."""

    manager_interface = IButtons
    prefix = 'buttons'

    def __init__(self, *args):
        buttons = []
        for arg in args:
            if IInterface.providedBy(arg):
                for name, button in getFieldsInOrder(arg):
                    if IButton.providedBy(button):
                        buttons.append((name, button))
            elif self.manager_interface.providedBy(arg):
                buttons += arg.items()
            elif IButton.providedBy(arg):
                if not arg.__name__:
                    arg.__name__ = create_id(arg.title)
                buttons.append((arg.__name__, arg))
            else:
                raise TypeError("Unrecognized argument type", arg)
        super().__init__(buttons)


@implementer(IButtonHandlers)
class Handlers:
    """Action Handlers for a Button-based form."""

    def __init__(self):
        self._registry = AdapterRegistry()
        self._handlers = ()

    def add_handler(self, button, handler):  # pylint: disable=redefined-outer-name
        """See interfaces.button.IButtonHandlers"""
        # Create a specification for the button
        button_spec = get_specification(button)
        if isinstance(button_spec, class_types):
            button_spec = implementedBy(button_spec)
        # Register the handler
        self._registry.register(
            (button_spec,), IButtonHandler, '', handler)
        self._handlers += ((button, handler),)

    def get_handler(self, button):
        """See interfaces.button.IButtonHandlers"""
        button_provided = providedBy(button)
        # pylint: disable=no-member
        return self._registry.lookup1(button_provided, IButtonHandler)

    def copy(self):
        """See interfaces.button.IButtonHandlers"""
        handlers = Handlers()
        for button, handler in self._handlers:  # pylint: disable=redefined-outer-name
            handlers.add_handler(button, handler)
        return handlers

    def __add__(self, other):
        """See interfaces.button.IButtonHandlers"""
        if not isinstance(other, Handlers):
            raise NotImplementedError
        handlers = self.copy()
        for button, handler in other._handlers:  # pylint: disable=redefined-outer-name
            handlers.add_handler(button, handler)
        return handlers

    def __repr__(self):
        # pylint: disable=redefined-outer-name
        return '<Handlers %r>' % [handler for button, handler in self._handlers]


@implementer(IButtonHandler)
class Handler:
    """Button handler class"""

    def __init__(self, button, func):
        self.button = button
        self.func = func

    def __call__(self, form, action):
        return self.func(form, action)

    def __repr__(self):
        return '<%s for %r>' %(self.__class__.__name__, self.button)


def handler(button):
    """A decorator for defining a success handler."""

    def create_handler(func):
        handler = Handler(button, func)  # pylint: disable=redefined-outer-name
        frame = sys._getframe(1)  # pylint: disable=protected-access
        f_locals = frame.f_locals
        handlers = f_locals.setdefault('handlers', Handlers())
        handlers.add_handler(button, handler)
        return handler

    return create_handler


def button_and_handler(title, **kwargs):
    """Button with handler method decorator"""
    # Add the title to button constructor keyword arguments
    kwargs['title'] = title
    # Extract directly provided interfaces:
    provides = kwargs.pop('provides', ())
    # Create button and add it to the button manager
    button = Button(**kwargs)
    alsoProvides(button, provides)
    frame = sys._getframe(1)  # pylint: disable=protected-access
    f_locals = frame.f_locals
    f_locals.setdefault('buttons', Buttons())
    f_locals['buttons'] += Buttons(button)
    # Return the handler decorator
    return handler(button)


@adapter_config(required=(IFormLayer, IButton),
                provides=IButtonAction)
class ButtonAction(Action, SubmitWidget, Location):
    """Button action"""

    def __init__(self, request, field):
        Action.__init__(self, request, field.title)
        SubmitWidget.__init__(self, request)
        self.field = field

    @property
    def access_key(self):
        """Button access key"""
        return self.field.access_key

    @property
    def value(self):
        """Button value"""
        return self.title

    @property
    def id(self):
        """Button ID"""
        return self.name.replace('.', '-')


@adapter_config(required=(IFormLayer, IImageButton),
                provides=IButtonAction)
class ImageButtonAction(ImageWidget, ButtonAction):
    """Image button action"""

    def __init__(self, request, field):  # pylint: disable=super-init-not-called
        Action.__init__(self, request, field.title)  # pylint: disable=non-parent-init-called
        SubmitWidget.__init__(self, request)  # pylint: disable=non-parent-init-called
        self.field = field

    @property
    def src(self):
        """Image source"""
        return to_unicode(self.field.image)

    def is_executed(self):
        return self.name + '.x' in self.request.params


@adapter_config(required=(IButtonForm, Interface, Interface),
                provides=IActions)
class ButtonActions(Actions):
    """Button actions manager"""

    def update(self):
        """See pyams_form.interfaces.button.IActions."""
        # Create a unique prefix.
        prefix = expand_prefix(self.form.prefix)
        prefix += expand_prefix(self.form.buttons.prefix)
        # Walk through each field, making an action out of it.
        d = {}
        d.update(self)
        registry = self.request.registry
        for name, button in self.form.buttons.items():
            # Step 1: Only create an action for the button, if the condition is
            #         fulfilled.
            if button.condition is not None and not button.condition(self.form):
                # Step 1.1: If the action already existed, but now the
                #           condition became false, remove the old action.
                if name in d:
                    del d[name]
                continue
            # Step 2: Get the action for the given button.
            new_button = True
            if name in self:
                button_action = self[name]
                new_button = False
            elif button.action_factory is not None:
                button_action = button.action_factory(self.request, button)
            else:
                button_action = registry.getMultiAdapter((self.request, button), IButtonAction)
            # Step 3: Set the name on the button
            button_action.name = prefix + name
            # Step 4: Set any custom attribute values.
            title = registry.queryMultiAdapter((self.form, self.request, self.content,
                                                button, self),
                                               IValue, name='title')
            if title is not None:
                button_action.title = title.get()
            # Step 5: Set the form
            button_action.form = self.form
            if not IFormAware.providedBy(button_action):
                alsoProvides(button_action, IFormAware)
            # Step 6: Update the new action
            button_action.update()
            registry.notify(AfterWidgetUpdateEvent(button_action))
            # Step 7: Add the widget to the manager
            if new_button:
                d[name] = button_action
                locate(button_action, self, name)

        self.create_according_to_list(d, self.form.buttons.keys())


@adapter_config(required=(IHandlerForm, Interface, Interface, ButtonAction),
                provides=IActionHandler)
class ButtonActionHandler(ActionHandlerBase):
    """Button action handler"""

    def __call__(self):
        # pylint: disable=redefined-outer-name
        handler = self.form.handlers.get_handler(self.action.field)
        # If no handler is found, then that's okay too.
        if handler is None:
            return None
        return handler(self.form, self.action)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.error module

This module provides snippets to display error messages.
"""

from zope.component import adapter
from zope.interface import Invalid, implementer
from zope.schema import ValidationError

from pyams_form.interfaces import IValue
from pyams_form.interfaces.error import IErrorViewSnippet, IMultipleErrors
from pyams_form.util import get_specification
from pyams_form.value import ComputedValueCreator, StaticValueCreator
from pyams_template.template import get_view_template
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


# pylint: disable=invalid-name
ErrorViewMessage = StaticValueCreator(
    discriminators=('error', 'request', 'widget', 'field', 'form', 'content')
)

# pylint: disable=invalid-name
ComputedErrorViewMessage = ComputedValueCreator(
    discriminators=('error', 'request', 'widget', 'field', 'form', 'content')
)


def ErrorViewDiscriminators(error_view, error=None, request=None, widget=None,
                            field=None, form=None, content=None):
    # pylint: disable=invalid-name,too-many-arguments
    """Error view discriminators"""
    adapter(get_specification(error),
            get_specification(request),
            get_specification(widget),
            get_specification(field),
            get_specification(form),
            get_specification(content))(error_view)


@adapter_config(required=(ValidationError, None, None, None, None, None),
                provides=IErrorViewSnippet)
@implementer(IErrorViewSnippet)
class ErrorViewSnippet:  # pylint: disable=too-many-instance-attributes
    """Base error view snippet."""

    message = None

    def __init__(self, error, request, widget, field, form, content):
        # pylint: disable=too-many-arguments
        self.error = self.context = error
        self.request = request
        self.widget = widget
        self.field = field
        self.form = form
        self.content = content

    def create_message(self):
        """Get error message"""
        return self.error.doc()

    def update(self):
        """Update snippet content"""
        registry = self.request.registry
        value = registry.queryMultiAdapter((self.context, self.request, self.widget,
                                            self.field, self.form, self.content),
                                           IValue, name='message')
        if value is not None:
            self.message = value.get()
        else:
            self.message = self.create_message()

    render = get_view_template()

    def __repr__(self):
        return '<%s for %s>' % (self.__class__.__name__, self.error.__class__.__name__)


@adapter_config(required=(ValueError, None, None, None, None, None),
                provides=IErrorViewSnippet)
class ValueErrorViewSnippet(ErrorViewSnippet):
    """An error view for ValueError."""

    default_message = _('The system could not process the given value.')

    def create_message(self):
        return self.default_message


@adapter_config(required=(Invalid, None, None, None, None, None),
                provides=IErrorViewSnippet)
class InvalidErrorViewSnippet(ErrorViewSnippet):
    """Error view snippet."""

    def create_message(self):
        return self.error.args[0]


@implementer(IMultipleErrors)
class MultipleErrors(Exception):
    """An error that contains many errors"""

    def __init__(self, errors):  # pylint: disable=super-init-not-called
        self.errors = errors


@adapter_config(required=(IMultipleErrors, None, None, None, None, None),
                provides=IErrorViewSnippet)
class MultipleErrorViewSnippet(ErrorViewSnippet):
    """Error view snippet for multiple errors."""

    def update(self):
        """Snippet update"""

    def render(self):
        """Render multiple errors"""
        return ''.join([view.render() for view in self.error.errors])



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.interfaces.widget module

This module include all widget-related interfaces (except IWidgetLayoutTemplate interface
which is defined into main interfaces module to avoid cyclic dependencies.
"""

from zope.interface import Attribute, Interface
from zope.location import ILocation
from zope.schema import ASCIILine, Bool, Field, Object, Text, TextLine, Tuple

from pyams_form.interfaces import DISPLAY_MODE, HIDDEN_MODE, IManager, INPUT_MODE, ITerms
from pyams_form.template import widget_layout_config, widget_template_config
from pyams_layer.interfaces import IFormLayer
from pyams_utils.interfaces.form import NO_VALUE


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


class IFieldWidget(Interface):
    """Offers a field attribute.

    For advanced uses the widget will make decisions based on the field
    it is rendered for.
    """

    field = Field(title=_('Field'),
                  description=_('The schema field which the widget is representing'),
                  required=True)


@widget_layout_config(mode=INPUT_MODE,
                      layer=IFormLayer,
                      template='templates/widget-layout.pt')
@widget_layout_config(mode=DISPLAY_MODE,
                      layer=IFormLayer,
                      template='templates/widget-layout-display.pt')
@widget_layout_config(mode=HIDDEN_MODE,
                      layer=IFormLayer,
                      template='templates/widget-layout-hidden.pt')
class IWidget(ILocation):
    """A widget within a form"""

    name = ASCIILine(title=_('Name'),
                     description=_('The name the widget is known under'),
                     required=True)

    label = TextLine(title=_('Label'),
                     description=_('''The widget label.

        Label may be translated for the request.

        The attribute may be implemented as either a read-write or read-only
        property, depending on the requirements for a specific implementation.
        '''),
                     required=True)

    mode = ASCIILine(title=_('Mode'),
                     description=_('A widget mode'),
                     default=INPUT_MODE,
                     required=True)

    required = Bool(title=_('Required'),
                    description=_('If true the widget should be displayed as required '
                                  'input'),
                    default=False,
                    required=True)

    error = Field(title=_('Error'),
                  description=_('If an error occurred during any step, the error view '
                                'stored here'),
                  required=False)

    value = Field(title=_('Value'),
                  description=_('The value that the widget represents'),
                  required=False)

    template = Attribute('''The widget template''')
    layout = Attribute('''The widget layout template''')

    ignore_request = Bool(title=_('Ignore Request'),
                          description=_('A flag, when set, forces the widget not to look at '
                                        'the request for a value'),
                          default=False,
                          required=False)

    # ugly thing to remove set_errors parameter from extract
    set_errors = Bool(title=_('Set errors'),
                      description=_('A flag, when set, the widget sets error messages '
                                    'on calling extract()'),
                      default=True,
                      required=False)

    # a bit different from ignore_required_on_extract, because we record
    # here the fact, but for IValidator, because the check happens there
    ignore_required_on_validation = Bool(title=_('Ignore Required validation'),
                                         description=_(
                                             "If set then required fields will pass validation "
                                             "regardless whether they're filled in or not"),
                                         default=False,
                                         required=True)

    show_default = Bool(title=_('Show default value'),
                        description=_('A flag, when set, makes the widget to display '
                                      'field|adapter provided default values'),
                        default=True,
                        required=False)

    def extract(self, default=NO_VALUE):
        """Extract the string value(s) of the widget from the form.

        The return value may be any Python construct, but is typically a
        simple string, sequence of strings or a dictionary.

        The value *must not* be converted into a native format.

        If an error occurs during the extraction, the default value should be
        returned. Since this should never happen, if the widget is properly
        designed and used, it is okay to NOT raise an error here, since we do
        not want to crash the system during an inproper request.

        If there is no value to extract, the default is to be returned.
        """

    def update(self):
        """Setup all of the widget information used for displaying."""

    def render(self):
        """Render the plain widget without additional layout"""

    def json_data(self):
        """Returns a dictionary for the widget"""

    def __call__(self):  # pylint: disable=signature-differs
        """Render a layout template which is calling widget/render"""


class ISequenceWidget(IWidget):
    """Term based sequence widget base.

    The sequence widget is used for select items from a sequence. Don't get
    confused, this widget does support to choose one or more values from a
    sequence. The word sequence is not used for the schema field, it's used
    for the values where this widget can choose from.

    This widget base class is used for build single or sequence values based
    on a sequence which is in most use case a collection. e.g.
    IList of IChoice for sequence values or IChoice for single values.

    See also the MultiWidget for build sequence values based on none collection
    based values. e.g. IList of ITextLine
    """

    no_value_token = ASCIILine(title=_('NO_VALUE Token'),
                               description=_(
                                   'The token to be used, if no value has been selected.'))

    terms = Object(title=_('Terms'),
                   description=_('A component that provides the options for selection'),
                   schema=ITerms)

    def update_terms(self):
        """Update the widget's ``terms`` attribute and return the terms.

        This method can be used by external components to get the terms
        without having to worry whether they are already created or not.
        """


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/multi-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/multi-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/multi-hidden.pt')
class IMultiWidget(IWidget):
    """None Term based sequence widget base.

    The multi widget is used for ITuple, IList or IDict if no other widget is defined.

    Some IList or ITuple are using another specialized widget if they can
    choose from a collection. e.g. a IList of IChoice. The base class of such
    widget is the ISequenceWidget.

    This widget can handle none collection based sequences and offers add or
    remove values to or from the sequence. Each sequence value get rendered by
    it's own relevant widget. e.g. IList of ITextLine or ITuple of IInt
    """


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/select-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/select-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/select-hidden.pt')
class ISelectWidget(ISequenceWidget):
    """Select widget with ITerms option."""

    prompt = Bool(title=_('Prompt'),
                  description=_('A flag, when set, enables a choice explicitely '
                                'requesting the user to choose a value'),
                  default=False)

    items = Tuple(title=_('Items'),
                  description=_('A collection of dictionaries containing all pieces of '
                                'information for rendering. The following keys must '
                                'be in each dictionary: id, value, content, selected'))

    no_value_message = Text(title=_('No-Value Message'),
                            description=_('A human-readable text that is displayed to refer the '
                                          'missing value.'))

    prompt_message = Text(title=_('Prompt Message'),
                          description=_('A human-readable text that is displayed to refer the '
                                        'missing value.'))


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/orderedselect-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/orderedselect-display.pt')
class IOrderedSelectWidget(ISequenceWidget):
    """Ordered Select widget with ITerms option."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/checkbox-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/checkbox-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/checkbox-hidden.pt')
class ICheckBoxWidget(ISequenceWidget):
    """Checbox widget."""


class ISingleCheckBoxWidget(ICheckBoxWidget):
    """Single Checbox widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/radio-input.pt')
@widget_template_config(mode='{}-single'.format(INPUT_MODE),
                        layer=IFormLayer,
                        template='templates/radio-input-single.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/radio-display.pt')
@widget_template_config(mode='{}-single'.format(DISPLAY_MODE),
                        layer=IFormLayer,
                        template='templates/radio-display-single.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/radio-hidden.pt')
@widget_template_config(mode='{}-single'.format(HIDDEN_MODE),
                        layer=IFormLayer,
                        template='templates/radio-hidden-single.pt')
class IRadioWidget(ISequenceWidget):
    """Radio widget."""

    def render_for_value(self, value):
        """Render a single radio button element for a given value.

        Here the word ``value`` is used in the HTML sense, in other
        words it is a term token.
        """


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/submit-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/submit-display.pt')
class ISubmitWidget(IWidget):
    """Submit widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/image-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/image-display.pt')
class IImageWidget(IWidget):
    """Submit widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/button-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/button-display.pt')
class IButtonWidget(IWidget):
    """Button widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/textarea-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/textarea-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/textarea-hidden.pt')
class ITextAreaWidget(IWidget):
    """Text widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/textlines-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/textlines-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/textlines-hidden.pt')
class ITextLinesWidget(IWidget):
    """Text lines widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/text-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/text-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/text-hidden.pt')
class ITextWidget(IWidget):
    """Text widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/file-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/file-display.pt')
class IFileWidget(ITextWidget):
    """File widget."""


class IMediaFileWidget(IFileWidget):
    """Media file widget marker interface"""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/password-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/password-display.pt')
class IPasswordWidget(ITextWidget):
    """Password widget."""


@widget_template_config(mode=INPUT_MODE,
                        layer=IFormLayer,
                        template='templates/object-input.pt')
@widget_template_config(mode=DISPLAY_MODE,
                        layer=IFormLayer,
                        template='templates/object-display.pt')
@widget_template_config(mode=HIDDEN_MODE,
                        layer=IFormLayer,
                        template='templates/object-hidden.pt')
class IObjectWidget(IWidget):
    """Object widget."""

    def setup_fields(self):
        """setup fields on the widget, by default taking the fields of
        self.schema"""


class IWidgets(IManager):
    """A widget manager"""

    prefix = ASCIILine(title=_('Prefix'),
                       description=_('The prefix of the widgets'),
                       default='widgets.',
                       required=True)

    mode = ASCIILine(title=_('Prefix'),
                     description=_('The prefix of the widgets'),
                     default=INPUT_MODE,
                     required=True)

    errors = Field(title=_('Errors'),
                   description=_('The collection of errors that occured during '
                                 'validation'),
                   default=(),
                   required=True)

    ignore_context = Bool(title=_('Ignore Context'),
                          description=_('If set the context is ignored to retrieve a value'),
                          default=False,
                          required=True)

    ignore_request = Bool(title=_('Ignore Request'),
                          description=_('If set the request is ignored to retrieve a value'),
                          default=False,
                          required=True)

    ignore_readonly = Bool(title=_('Ignore Readonly'),
                           description=_('If set then readonly fields will also be shown'),
                           default=False,
                           required=True)

    ignore_required_on_extract = Bool(title=_('Ignore Required validation on extract'),
                                      description=_(
                                          "If set then required fields will pass validation "
                                          "on extract regardless whether they're filled in or not"),
                                      default=False,
                                      required=True)

    has_required_rields = Bool(title=_('Has required fields'),
                               description=_('A flag set when at least one field is marked as '
                                             'required'),
                               default=False,
                               required=False)

    # ugly thing to remove set_errors parameter from extract
    set_errors = Bool(title=_('Set errors'),
                      description=_('A flag, when set, the contained widgets set error '
                                    'messages on calling extract()'),
                      default=True,
                      required=False)

    def update(self):
        """Setup widgets."""

    def extract(self):
        """Extract the values from the widgets and validate them.
        """

    def extract_raw(self):
        """Extract the RAW/string values from the widgets and validate them.
        """


class IWidgetEvent(Interface):
    """A simple widget event."""

    widget = Object(title=_('Widget'),
                    description=_('The widget for which the event was created.'),
                    schema=IWidget)


class IAfterWidgetUpdateEvent(IWidgetEvent):
    """An event sent out after the widget was updated."""



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.validator module

Default validation adapters.
"""

import copy

from zope.component import adapter
from zope.interface import Interface, Invalid, alsoProvides, implementer
from zope.interface.interfaces import IInterface, IMethod
from zope.schema.interfaces import IBytes, IField

from pyams_form.interfaces import IData, IManagerValidator, IValidator, IValue
from pyams_form.interfaces.form import IContextAware
from pyams_form.interfaces.widget import IFileWidget
from pyams_form.util import changed_widget, get_specification
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import IDataManager, NOT_CHANGED, NO_VALUE
from pyams_utils.registry import get_current_registry


__docformat__ = 'restructuredtext'


@implementer(IValidator)
class StrictSimpleFieldValidator:
    """Strict simple field validator

    Validates all incoming values
    """

    def __init__(self, context, request, view, field, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.view = view
        self.field = field
        self.widget = widget

    def validate(self, value, force=False):  # pylint: disable=unused-argument
        """See interfaces.IValidator"""
        context = self.context
        field = self.field
        widget = self.widget
        if field.required and widget and widget.ignore_required_on_validation:
            # make the field not-required while checking
            field = copy.copy(field)
            field.required = False
        if context is not None:
            field = field.bind(context)
        registry = self.request.registry
        if value is NOT_CHANGED:
            if IContextAware.providedBy(widget) and not widget.ignore_context:
                # get value from context
                value = registry.getMultiAdapter((context, field), IDataManager).query()
            else:
                value = NO_VALUE
            if value is NO_VALUE:
                # look up default value
                value = field.default
                # pylint: disable=redefined-outer-name
                adapter = registry.queryMultiAdapter(
                    (context, self.request, self.view, field, widget),
                    IValue, name='default')
                if adapter:
                    value = adapter.get()
        return field.validate(value)

    def __repr__(self):
        return "<%s for %s['%s']>" % (
            self.__class__.__name__,
            self.field.interface.getName(),
            self.field.__name__)


@adapter_config(required=(Interface, Interface, Interface, IField, Interface),
                provides=IValidator)
class SimpleFieldValidator(StrictSimpleFieldValidator):
    """Simple Field Validator

    Ignores unchanged values.
    """

    def validate(self, value, force=False):
        """See interfaces.IValidator"""
        if value is self.field.missing_value:
            # let missing values run into stricter validation
            # most important case is not let required fields pass
            return super().validate(value, force)

        if not force:
            if value is NOT_CHANGED:
                # no need to validate unchanged values
                return None

            if self.widget and not changed_widget(
                    self.widget, value, field=self.field, context=self.context):
                # if new value == old value, no need to validate
                return None

        # otherwise StrictSimpleFieldValidator will do the job
        return super().validate(value, force)


@adapter_config(required=(Interface, Interface, Interface, IBytes, IFileWidget),
                provides=IValidator)
class FileUploadValidator(StrictSimpleFieldValidator):
    """File upload validator
    """
    # only FileUploadDataConverter seems to use NOT_CHANGED, but that needs
    # to be validated, because file upload is a special case
    # the most special case if when an ad-hoc IBytes field is required


def WidgetValidatorDiscriminators(validator, context=None, request=None, view=None,
                                  field=None, widget=None):
    # pylint: disable=invalid-name,too-many-arguments
    """Widget validator discriminators"""
    adapter(
        get_specification(context),
        get_specification(request),
        get_specification(view),
        get_specification(field),
        get_specification(widget)
    )(validator)


class NoInputData(Invalid):
    """There was no input data because:

    - It wasn't asked for

    - It wasn't entered by the user

    - It was entered by the user, but the value entered was invalid

    This exception is part of the internal implementation of checkInvariants.

    """


@implementer(IData)
class Data:
    """Form data proxy implementation"""

    def __init__(self, schema, data, context):
        self._Data_data___ = data  # pylint: disable=invalid-name
        self._Data_schema___ = schema  # pylint: disable=invalid-name
        alsoProvides(self, schema)
        self.__context__ = context

    def __getattr__(self, name):
        schema = self._Data_schema___
        data = self._Data_data___
        try:
            field = schema[name]
        except KeyError as err:
            raise AttributeError(name) from err
        # If the found field is a method, then raise an error.
        if IMethod.providedBy(field):
            raise RuntimeError("Data value is not a schema field", name)
        # Try to get the value for the field
        value = data.get(name, data)
        if value is data:
            if self.__context__ is None:
                raise NoInputData(name)
            registry = get_current_registry()
            dman = registry.getMultiAdapter((self.__context__, field), IDataManager)
            value = dman.get()
        # Optimization: Once we know we have a good value, set it as an
        # attribute for faster access.
        setattr(self, name, value)
        return value


@adapter_config(required=(Interface, Interface, Interface, IInterface, Interface),
                provides=IManagerValidator)
class InvariantsValidator:
    """Simple Field Validator"""

    def __init__(self, context, request, view, schema, manager):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.view = view
        self.schema = schema
        self.manager = manager

    def validate(self, data):
        """See interfaces.IManagerValidator"""
        return self.validate_object(Data(self.schema, data, self.context))

    def validate_object(self, object):  # pylint: disable=redefined-builtin
        """validate given object"""
        errors = []
        try:
            self.schema.validateInvariants(object, errors)
        except Invalid:
            pass  # Just collect the errors

        return tuple([error for error in errors
                      if not isinstance(error, NoInputData)])

    def __repr__(self):
        return '<%s for %s>' %(self.__class__.__name__, self.schema.getName())


def WidgetsValidatorDiscriminators(validator, context=None, request=None, view=None,
                                   schema=None, manager=None):
    # pylint: disable=invalid-name,too-many-arguments
    """Widgets validator discriminators"""
    adapter(
        get_specification(context),
        get_specification(request),
        get_specification(view),
        get_specification(schema),
        get_specification(manager)
    )(validator)
